#!/usr/bin/env ruby
#
# addimage2mp3.rb - Script for embedding album cover images in MP3 files.
#                   This requires an existing ID3-V2 tag in the file.
#
# (c) 2008 Budi Rachmanto   <rac AT are.ma>     v0.7.6
#     2005 Mark Kretschmann <markey AT web.de>  v0.7
# License: GNU General Public License V2

# Unsynchronize the entire ID3-V2 tag, frame by frame (replace 0xfff* with 0xff00f*)
def unsynchronize( data )
  data[5] |= 0b10000000  # Set Unsychronization global tag flag
  index = 10  # Skip ID3 Header

  until data[index] == 0 or data[index..index+2] == "3DI"
    frametype = data[index..index+3]
    framesize = data[index+4]*2**21 + data[index+5]*2**14 + data[index+6]*2**7 + data[index+7]

    # Check if frame is already unsynced
    if data[index+9] & 0x02 == 0
      puts( "#{frametype}  unsynced " )
      data[index+9] |= 0x02
      index += 10

      framesize.times() do
        sync = data[index] == 0xff
        sync1 = data[index+1] & 0b11100000 == 0b11100000
        sync2 = data[index+1] & 0b11111111 == 0b00000000

        if sync and ( sync1 or sync2 ) and index < framesize - 2
          print( "." ) if sync1
          print( "O" ) if sync2
          data[index+1, 0] = "\0"
          framesize += 1
          index += 1
        end

        index += 1
      end
    else
      puts( "#{frametype}  synced" )
      index += framesize + 10
    end
  end

  # Index == new tag length
  return index - 10
end

# Returns the size of the ID3-V2 tag. In other words, the offset from
# where the real mp3 data starts.
# @see http://id3lib.sourceforge.net/id3/id3v2com-00.html#sec3.1
def getId3v2Size( data )
  return data[6]*2**21 + data[7]*2**14 + data[8]*2**7 + data[9]
end

# Sets the size of the ID3-V2 tag.
# @see http://id3lib.sourceforge.net/id3/id3v2com-00.html#sec3.1
def setId3v2Size( data, size )
  data[6] = size >> 21 & 0x7f
  data[7] = size >> 14 & 0x7f
  data[8] = size >> 7  & 0x7f
  data[9] = size >> 0  & 0x7f
end

############################################################################
# FRAME HANDLERS
############################################################################

def getId3v2FrameSize( vmajor, data, idx )
  fsize = case vmajor
#   when 2 then data[idx+3]*2**16 + data[idx+4]*2**8  + data[idx+7]
    when 3 then data[idx+4]*2**24 + data[idx+5]*2**16 + data[idx+6]*2**8 + data[idx+7]
    when 4 then data[idx+4]*2**21 + data[idx+5]*2**14 + data[idx+6]*2**7 + data[idx+7]
    else
      puts( "Error: Major number %02x is not supported." % [vmajor] )
      exit( 1 )
    end
  return fsize
end

def setId3v24FrameSize( data, idx, size )
  data[idx+4] = size >> 21 & 0x7f
  data[idx+5] = size >> 14 & 0x7f
  data[idx+6] = size >> 7  & 0x7f
  data[idx+7] = size >> 0  & 0x7f
end

############################################################################
# MAIN
############################################################################

path        = ""
destination = ""

if $*.length() < 2 or $*[0] == "--help"
  puts( "Usage: addimage2mp3.rb image source [destination]" )
  puts()
  puts( "Adds an image to the ID3-V2 tag of a MP3 file." )
  puts( "Requires an existing ID3-V2 tag inthe file." )
  exit( 1 )
end

imagepath   = $*[0]
path        = $*[1]
destination = path
if $*.length() == 3
  destination = $*[2]
end

unless FileTest::readable_real?( imagepath )
  puts( "Error: Image not found or not readable." )
  exit( 1 )
end

unless FileTest::readable_real?( path )
  puts( "Error: Source not found or not readable." )
  exit( 1 )
end

unless File.extname( path ).downcase() == ".mp3"
  puts( "Error: File is not mp3." )
  exit( 1 )
end

mimetype = case File.extname( imagepath ).downcase()
  when ".bmp"  then "image/bmp"
  when ".gif"  then "image/gif"
  when ".jpg"  then "image/jpeg"
  when ".jpeg" then "image/jpeg"
  when ".png"  then "image/png"
  when ""      then "image/png"  # amaroK's cover images are always PNG
  else
    puts( "Error: Image type invalid." )
    exit( 1 )
end

file  = File.new( imagepath, "r" )
image = file.read()

apicframe = String.new()
apicframe << 0x00  # text encoding
apicframe << mimetype
apicframe << 0x00  # mimetype end
apicframe << 0x03  # Picture type (Cover front)
apicframe << 0x00  # Description (empty)
apicframe << image

apicheader = String.new()
apicheader << "APIC"
apicheader << ( ( apicframe.length() >> 21 ) & 0x7f )
apicheader << ( ( apicframe.length() >> 14 ) & 0x7f )
apicheader << ( ( apicframe.length() >> 7  ) & 0x7f )
apicheader << ( ( apicframe.length() >> 0  ) & 0x7f )
apicheader << 0x00  # Flags
apicheader << 0x00  # Flags

file = File.new( path, "r" )
data = file.read()

if data[0,3] == "ID3"
  id3length = getId3v2Size( data )
  vmajor    = data[3] # major number
  puts( "ID3-V2.#{vmajor} detected. Tag size: #{id3length}" )
else
  puts( "ID3-V1 detected." )
  puts( "Error: File does not have a ID3-V2 tag." )
  exit( 1 )
end

# Find end of last ID3 frame, before padding or footer (if present)
puts()
puts( "Locating end of last ID3 frame.." )

retrycnt = 0
begin
  data[3] = 0x04 # enforcedly change to Id3v2.4
  index   = 10   # Skip ID3 Header
  until data[index] == 0 or data[index..index+2] == "3DI" or index >= 10 + id3length
    vmajor = 3 if (data[index+4]>>7)+(data[index+5]>>7)+(data[index+6]>>7)+(data[index+7]>>7) > 0
    frametype = data[index..index+3]
    framesize = getId3v2FrameSize( vmajor, data, index )
    puts( "Frame Type:%02x%02x%02x%02x/%s, size:%02x%02x%02x%02x/%d" % [data[index], data[index+1], data[index+2], data[index+3], frametype, data[index+4], data[index+5], data[index+6], data[index+7], framesize] )
    setId3v24FrameSize( data, index, framesize ) # enforcedly change to Id3v2.4 frame size format
    if frametype == "APIC"
      puts( "Warning: Image found inside the file!" )
      pType = index + 11
      until data[pType] == 0
        pType += 1
      end
    end
    if frametype == "APIC" and ( data[pType+1] == 3 or data[pType+1] == 0 ) # Picture = "Cover front"/"other"
      puts( "Warning: File already had front cover image. OVERWRITING!" )
      origapicsize = 10 + framesize
      data[index, origapicsize] = ""
      id3length -= origapicsize
    else
      if ( index + framesize ) > id3length
        puts( "Warning: Wrong frame size!" )
        break
      else
        index += 10 + framesize # frameheader length = 10
      end
    end
  end
rescue
  puts()
  if vmajor == 3 and retrycnt == 0
    puts( "ID3-V2.3 found but the frame tag is wrong, maybe it is V2.4? Retrying..." )
    file = File.new( path, "r" )
    data = file.read()
    id3length = getId3v2Size( data )
    vmajor    = 4
    retrycnt += 1
    retry
  else
    puts( "Error: Broken frame tag, #{$!}" )
    exit( 1 )
  end
end

# Insert APIC frame into string, after the last ID3 frame
data[index, 0] = apicheader + apicframe
id3length += apicheader.length() + apicframe.length()

# Unsynchronization isn't supported by TagLib at this point :|
#
# puts()
# puts( "Unsynchronizing tag.." )
# id3length = unsynchronize( data )
# data[5] |= 0b10000000  # Set ID3 Unsychronization flag

# Adjust ID3V2 tag size
setId3v2Size( data, id3length )
puts()
puts( "Adjusting new ID3 size: #{id3length + 10}" )

puts()
print( "Writing file..  " )
begin
  destfile = File::open( destination, File::CREAT|File::TRUNC|File::WRONLY )
rescue
  puts( "Error: Destination file is not writable." )
  exit( 1 )
end

destfile << data
puts( "done." )

puts()
puts( "done." )
